import asyncio
from time import sleep
import websockets
import y_py as Y
import queue
import concurrent.futures
import threading

# Code based on the [`websockets` patter documentation](https://websockets.readthedocs.io/en/stable/howto/patterns.html)

class YDocWSClient:

    def __init__(self, uri = "ws://localhost:7654"):
        self.send_q = queue.Queue()
        self.recv_q = queue.Queue()
        self.uri = uri
        def async_loop():
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

            loop.run_until_complete(self.start_ws_client())
            loop.close()
        ws_thread = threading.Thread(target=async_loop, daemon=True)
        ws_thread.start()

    def send_updates(self, txn_event: Y.AfterTransactionEvent):
        update = txn_event.get_update()
        # Sometimes transactions don't write, which means updates are empty.
        # We only care about updates with meaningful mutations.
        if update != b'\x00\x00':
            self.send_q.put_nowait(update)

    def apply_updates(self, doc: Y.YDoc):
        while not self.recv_q.empty():
            update = self.recv_q.get_nowait()
            Y.apply_update(doc, update)
    
    async def client_handler(self, websocket):
        consumer_task = asyncio.create_task(self.consumer_handler(websocket))
        producer_task = asyncio.create_task(self.producer_handler(websocket))
        done, pending = await asyncio.wait(
            [consumer_task, producer_task],
            return_when=asyncio.FIRST_COMPLETED,
        )
        for task in pending:
            task.cancel()

    async def consumer_handler(self, websocket):
        async for message in websocket:
            self.recv_q.put_nowait(message)
    
    async def producer_handler(self, websocket):
        loop = asyncio.get_running_loop()
        while True:
            update = await loop.run_in_executor(None,self.send_q.get)
            await websocket.send(update)

    async def start_ws_client(self):
        async with websockets.connect(self.uri) as websocket:
            await self.client_handler(websocket)
    


from y_py import YDoc, AfterTransactionEvent

import y_py as Y
import pytest


def test_constructor_options():
    # Ensure that valid versions can be called without error
    YDoc()
    with_id = YDoc(1)
    assert with_id.client_id == 1
    YDoc(1, "utf-8", True)
    YDoc(client_id=2, offset_kind="utf-8", skip_gc=True)
    YDoc(client_id=3)
    YDoc(4, skip_gc=True)

    # Handle encoding string variation
    YDoc(offset_kind="utf8")
    YDoc(offset_kind="utf-8")
    YDoc(offset_kind="UTF-8")
    YDoc(offset_kind="UTF32")

    # Ensure that incorrect encodings throw error
    with pytest.raises(ValueError):
        YDoc(offset_kind="UTF-0xDEADBEEF")
    with pytest.raises(ValueError):
        YDoc(offset_kind="😬")


def test_encoding():
    """
    Tests encoding / decoding all primitive data types in an array.
    """
    doc = YDoc()
    receiver = YDoc()
    array = doc.get_array("test")
    contents = [True, 42, "string"]
    with doc.begin_transaction() as txn:
        array.insert_range(txn, 0, contents)

    state_vec = Y.encode_state_vector(receiver)
    update = Y.encode_state_as_update(doc, state_vec)
    Y.apply_update(receiver, update)
    value = list(receiver.get_array("test"))
    assert value == contents


def test_boolean_encoding():
    """
    Makes sure the boolean types are preserved.
    Added due to bug where bools turn to ints during encoding /decoding.
    """
    doc = YDoc()
    receiver = YDoc()
    array = doc.get_array("test")
    with doc.begin_transaction() as txn:
        array.insert(txn, 0, True)

    state_vec = Y.encode_state_vector(receiver)
    update = Y.encode_state_as_update(doc, state_vec)
    Y.apply_update(receiver, update)
    value = list(receiver.get_array("test"))
    assert type(value[0]) == type(True)


def test_tutorial():
    d1 = Y.YDoc()
    text = d1.get_text("test")
    with d1.begin_transaction() as txn:
        text.extend(txn, "hello world!")

    d2 = Y.YDoc()
    state_vector = Y.encode_state_vector(d2)
    diff = Y.encode_state_as_update(d1, state_vector)
    Y.apply_update(d2, diff)

    value = str(d2.get_text("test"))

    assert value == "hello world!"


def test_observe_after_transaction():
    doc = Y.YDoc()
    text = doc.get_text("test")
    before_state = None
    after_state = None
    delete_set = None

    def callback(event):
        nonlocal before_state
        nonlocal after_state
        nonlocal delete_set
        before_state = event.before_state
        after_state = event.after_state
        delete_set = event.delete_set

    # Subscribe callback
    sub = doc.observe_after_transaction(callback)

    # Update the document
    with doc.begin_transaction() as txn:
        text.insert(txn, 0, "abc")
        text.delete_range(txn, 1, 2)

    assert before_state != None
    assert after_state != None
    assert delete_set != None


def test_get_update():
    """
    Ensures that developers can access the encoded update data in the `observe_after_transaction` event.
    """
    d = Y.YDoc()
    m = d.get_map("foo")
    r = d.get_map("foo")
    update: bytes = None

    def get_update(event: AfterTransactionEvent) -> None:
        nonlocal update
        update = event.get_update()

    d.observe_after_transaction(get_update)

    with d.begin_transaction() as txn:
        m.set(txn, "hi", "there")

    assert type(update) == bytes



import json
from copy import deepcopy

import pytest
from test_helper import exchange_updates
from y_py import YArray, YArrayEvent, YDoc


def test_inserts():
    d1 = YDoc(1)
    x = d1.get_array("test")

    with d1.begin_transaction() as txn:
        x.insert_range(txn, 0, [1, 2.5, "hello", ["world"], True])

    with d1.begin_transaction() as txn:
        x.extend(txn, [{"key": "value"}])

    expected = [1, 2.5, "hello", ["world"], True, {"key": "value"}]

    assert list(x) == expected

    d2 = YDoc(2)
    x = d2.get_array("test")

    exchange_updates([d1, d2])

    assert list(x) == expected

    # Ensure that preliminary types can be inserted
    integrated_array = d1.get_array("prelim_container")
    inserted_prelim = YArray(["insert"])
    extended_prelim = YArray(["extend"])

    with d1.begin_transaction() as txn:
        integrated_array.insert(txn,0,inserted_prelim)
        integrated_array.extend(txn, [extended_prelim])
    values = [list(a) for a in integrated_array]
    assert values == [["insert"], ["extend"]]



def test_to_string():
    arr = YArray([7, "awesome", True, ["nested"], {"testing": "dicts"}])
    expected_str = "[7, 'awesome', True, ['nested'], {'testing': 'dicts'}]"
    assert str(arr) == expected_str
    assert arr.__repr__() == f"YArray({expected_str})"

def test_to_json():
    contents = [7, "awesome", True, ["nested"], {"testing": "dicts"}]
    doc = YDoc()
    prelim = YArray(deepcopy(contents))
    integrated = doc.get_array("arr")
    with doc.begin_transaction() as txn:
        integrated.extend(txn, contents)
    expected_json = '[7,"awesome",true,["nested"],{"testing":"dicts"}]'
    assert integrated.to_json() == expected_json
    assert prelim.to_json() == expected_json
    
    # ensure that it works with python json
    assert json.loads(integrated.to_json()) == contents

def test_inserts_nested():
    d1 = YDoc()
    x = d1.get_array("test")
    to_list = lambda arr : [list(x) if type(x) == YArray else x for x in arr]
    nested = YArray()
    d1.transact(lambda txn: nested.append(txn, "world"))
    d1.transact(lambda txn: x.insert_range(txn, 0, [1, 2, nested, 3, 4]))
    d1.transact(lambda txn: nested.insert(txn, 0, "hello"))

    expected = [1, 2, ["hello", "world"], 3, 4]
    assert to_list(nested) == ["hello", "world"]
    assert type(x[2]) == YArray
    assert to_list(x) == expected

    d2 = YDoc()
    x = d2.get_array("test")

    exchange_updates([d1, d2])

    assert to_list(x) == expected


def test_delete():
    d1 = YDoc(1)
    assert d1.client_id == 1
    x = d1.get_array("test")

    d1.transact(
        lambda txn: x.insert_range(
            txn,
            0,
            [1, 2, ["hello", "world"], {"user": "unknown"}, "I'm here too!", True],
        )
    )
    d1.transact(lambda txn: x.delete_range(txn, 1, 3))

    expected = [1, "I'm here too!", True]

    assert list(x) == expected

    with d1.begin_transaction() as txn:
        x.delete(txn, 1)

    assert list(x) == [1.0, True]
    with pytest.raises(IndexError):
        with d1.begin_transaction() as txn:
            x.delete(txn, 2)

    d2 = YDoc(2)
    x = d2.get_array("test")

    exchange_updates([d1, d2])

    assert list(x) == [1.0, True]


def test_get():
    d1 = YDoc()
    integrated = d1.get_array("test")
    prelim = YArray()

    d1.transact(lambda txn: integrated.insert_range(txn, 0, [1, 2, True]))
    d1.transact(lambda txn: integrated.insert_range(txn, 1, ["hello", "world"]))

    expected = [1, "hello", "world", 2, True]
    prelim = YArray(expected)

    for arr in [integrated, prelim]:
        # Forward indexing
        for i, expected_value in enumerate(expected):
            assert arr[i] == expected_value

        with pytest.raises(IndexError):
            arr[5]

        # Negative indexing
        for i, expected_value in enumerate(reversed(expected)):
            index = -(i + 1)
            assert arr[index] == expected_value

        with pytest.raises(IndexError):
            arr[-6]

        # Slices
        assert arr[0:] == expected
        assert arr[4:1:-1] == expected[4:1:-1]
        assert arr[::-1] == expected[::-1]


def test_iterator():
    d1 = YDoc()
    x = d1.get_array("test")

    with d1.begin_transaction() as txn:
        x.insert_range(txn, 0, [1, 2, 3])
    assert len(x) == 3
    i = 1.0
    # Test iteration
    for v in x:
        assert v == i
        i += 1.0
    # Test contains
    assert 2.0 in x


def test_borrow_mut_edge_case():
    """
    Tests for incorrect overlap in successive mutable borrows of YTransaction and YArray.
    """
    doc = YDoc()
    arr = doc.get_array("test")
    with doc.begin_transaction() as txn:
        arr.insert_range(txn, 0, [1, 2, 3])

    # Ensure multiple transactions can be called in a row with the same variable name `txn`
    with doc.begin_transaction() as txn:
        # Ensure that multiple mutable borrow functions can be called in a tight loop
        for i in range(2000):
            arr.insert_range(txn, 2, [1, 2, 3])
            arr.delete_range(txn, 0, 3)


def test_observer():
    d1 = YDoc()

    x = d1.get_array("test")

    target = None
    delta = None

    def callback(e: YArrayEvent):
        nonlocal target
        nonlocal delta
        target = e.target
        delta = e.delta

    subscription_id = x.observe(callback)

    # insert initial data to an empty YArray
    with d1.begin_transaction() as txn:
        x.insert_range(txn, 0, [1, 2, 3, 4])
    assert list(target) == list(x)
    assert delta == [{"insert": [1, 2, 3, 4]}]

    target = None
    delta = None

    # remove 2 items from the middle
    with d1.begin_transaction() as txn:
        x.delete_range(txn, 1, 2)
    assert list(target) == list(x)
    assert delta == [{"retain": 1}, {"delete": 2}]

    target = None
    delta = None

    # insert item in the middle
    with d1.begin_transaction() as txn:
        x.insert(txn, 1, 5)
    assert list(target) == list(x)
    assert delta == [{"retain": 1}, {"insert": [5]}]

    target = None
    delta = None

    # Cancel the observer and make sure that callback is no longer called
    x.unobserve(subscription_id)

    with d1.begin_transaction() as txn:
        x.insert(txn, 1, 6)

    assert target == None
    assert delta == None


def test_deep_observe():
    """
    Ensure that changes to elements inside the array trigger a callback.
    """
    ydoc = YDoc()
    container = ydoc.get_array("container")
    yarray = YArray([1, 2])
    with ydoc.begin_transaction() as txn:
        container.append(txn, yarray)

    events = None

    def callback(e: list):
        nonlocal events
        events = e

    sub = container.observe_deep(callback)
    with ydoc.begin_transaction() as txn:
        container[0].append(txn, 3)

    assert events != None

    # Ensure that observer unsubscribes
    events = None
    container.unobserve(sub)
    with ydoc.begin_transaction() as txn:
        container[0].append(txn, 4)

    assert events == None

def test_move_to():
    """
    Ensure that move_to works.
    """
    doc = YDoc()
    arr = doc.get_array('test')

    # Move 0 to 10
    with doc.begin_transaction() as t:
        arr.delete_range(t, 0, len(arr))
        arr.extend(t, [0,1,2,3,4,5,6,7,8,9])
    doc.transact(lambda t: arr.move_to(t, 0, 10))
    assert list(arr) == [1,2,3,4,5,6,7,8,9,0]

    # Move 9 to 0
    with doc.begin_transaction() as t:
        arr.delete_range(t, 0, len(arr))
        arr.extend(t, [0,1,2,3,4,5,6,7,8,9])
    doc.transact(lambda t: arr.move_to(t, 9, 0))
    assert list(arr) == [9,0,1,2,3,4,5,6,7,8]

    # Move 6 to 5
    with doc.begin_transaction() as t:
        arr.delete_range(t, 0, len(arr))
        arr.extend(t, [0,1,2,3,4,5,6,7,8,9])
    doc.transact(lambda t: arr.move_to(t, 6, 5))
    assert list(arr) == [0,1,2,3,4,6,5,7,8,9]

    # Move -1 to 5
    with doc.begin_transaction() as t:
        arr.delete_range(t, 0, len(arr))
        arr.extend(t, [0,1,2,3,4,5,6,7,8,9])
    with pytest.raises(Exception):
        doc.transact(lambda t: arr.move_to(t, -1, 5))

    # Move 0 to -5
    with doc.begin_transaction() as t:
        arr.delete_range(t, 0, len(arr))
        arr.extend(t, [0,1,2,3,4,5,6,7,8,9])
    with pytest.raises(Exception):
        doc.transact(lambda t: arr.move_to(t, 0, -5))

@pytest.mark.skip("move_range_to has failing debug assert in yrs 0.16 with this test")
def test_move_range_to():
    """
    Ensure that move_range_to works.
    """
    doc = YDoc()
    arr = doc.get_array('test')

    # Move 1-2 to 4
    with doc.begin_transaction() as t:
        arr.delete_range(t, 0, len(arr))
        arr.extend(t, [0,1,2,3])
    doc.transact(lambda t: arr.move_range_to(t, 1, 2, 4))
    assert list(arr) == [0,3,1,2]

    # Move 0-0 to 10
    with doc.begin_transaction() as t:
        arr.delete_range(t, 0, len(arr))
        arr.extend(t, [0,1,2,3,4,5,6,7,8,9])
    doc.transact(lambda t: arr.move_range_to(t, 0, 0, 10))
    assert list(arr) == [1,2,3,4,5,6,7,8,9,0]

    # Move 0-1 to 10
    with doc.begin_transaction() as t:
        arr.delete_range(t, 0, len(arr))
        arr.extend(t, [0,1,2,3,4,5,6,7,8,9])
    doc.transact(lambda t: arr.move_range_to(t, 0, 1, 10))
    assert list(arr) == [2,3,4,5,6,7,8,9,0,1]


    # Move 3-5 to 7
    with doc.begin_transaction() as t:
        arr.delete_range(t, 0, len(arr))
        arr.extend(t, [0,1,2,3,4,5,6,7,8,9])
    doc.transact(lambda t: arr.move_range_to(t, 3, 5, 7))
    assert list(arr) == [0,1,2,6,3,4,5,7,8,9]

    # Move 1-0 to 10
    with doc.begin_transaction() as t:
        arr.delete_range(t, 0, len(arr))
        arr.extend(t, [0,1,2,3,4,5,6,7,8,9])
    doc.transact(lambda t: arr.move_range_to(t, 1, 0, 10))
    assert list(arr) == [0,1,2,3,4,5,6,7,8,9]

    # Move 3-5 to 5
    with doc.begin_transaction() as t:
        arr.delete_range(t, 0, len(arr))
        arr.extend(t, [0,1,2,3,4,5,6,7,8,9])
    doc.transact(lambda t: arr.move_range_to(t, 3, 5, 5))
    assert list(arr) == [0,1,2,3,4,5,6,7,8,9]

    # Move 9-9 to 0
    with doc.begin_transaction() as t:
        arr.delete_range(t, 0, len(arr))
        arr.extend(t, [0,1,2,3,4,5,6,7,8,9])
    doc.transact(lambda t: arr.move_range_to(t, 9, 9, 0))
    assert list(arr) == [9,0,1,2,3,4,5,6,7,8]

    # Move 8-9 to 0
    with doc.begin_transaction() as t:
        arr.delete_range(t, 0, len(arr))
        arr.extend(t, [0,1,2,3,4,5,6,7,8,9])
    doc.transact(lambda t: arr.move_range_to(t, 8, 9, 0))
    assert list(arr) == [8,9,0,1,2,3,4,5,6,7]

    # Move 4-6 to 3
    with doc.begin_transaction() as t:
        arr.delete_range(t, 0, len(arr))
        arr.extend(t, [0,1,2,3,4,5,6,7,8,9])
    doc.transact(lambda t: arr.move_range_to(t, 4, 6, 3))
    assert list(arr) == [0,1,2,4,5,6,3,7,8,9]

    # Move 3-5 to 3
    with doc.begin_transaction() as t:
        arr.delete_range(t, 0, len(arr))
        arr.extend(t, [0,1,2,3,4,5,6,7,8,9])
    doc.transact(lambda t: arr.move_range_to(t, 3, 5, 3))
    assert list(arr) == [0,1,2,3,4,5,6,7,8,9]

    # Move -1-2 to 5
    with doc.begin_transaction() as t:
        arr.delete_range(t, 0, len(arr))
        arr.extend(t, [0,1,2,3,4,5,6,7,8,9])
    with pytest.raises(Exception):
        doc.transact(lambda t: arr.move_range_to(t, -1, 2, 5))

    # Move 0--1 to 3
    with doc.begin_transaction() as t:
        arr.delete_range(t, 0, len(arr))
        arr.extend(t, [0,1,2,3,4,5,6,7,8,9])
    with pytest.raises(Exception):
        doc.transact(lambda t: arr.move_range_to(t, 0, -1, 3))



import json
from copy import deepcopy

import y_py as Y

# This is the original Lexical data
EXPECTED_JSON = json.loads(
    '{"root": {"__dir": "ltr", "__format": "", "__indent": 0, "__type": "root", "__version": 1, "children": [{"__type": "recommendation-wrapper", "__format": 0, "__indent": 0, "__dir": "ltr", "__kind": "write", "children": [{"__type": "paragraph", "__format": 0, "__indent": 0, "__dir": "ltr", "__textFormat": 0, "children": [{"__type": "text", "__format": 0, "__style": "", "__mode": 0, "__detail": 0, "text": "A"}]}, {"__type": "recommendation-buttons", "__kind": "write", "__recommendationDuid": "asdf1234asdf"}]}]}}'
)

# This is the Y update representation of the Lexical data
#
# YDoc loaded from Lexical data is represented internally as a tree of YXmlElement, YXmlText, YMap and SplittableString elements. All leaf nodes are always built with using YMap and SplittableString that are used to represent node attributes and text content respectively:
# YXmlElement
#     YXmlText
#         YMap             |___ Text leaf node
#         SplittableString |
#         YMap             |___ Text leaf node
#         SplittableString |
#     YXmlText
#         YXmlText
#             YMap             |___ Text leaf node
#             SplittableString |
#         YXmlText
#             YXmlText
#                 YXmlText
#                     YMap             |___ Text leaf node
#                     SplittableString |
#     YXmlText
#         YMap             |___ Text leaf node
#         SplittableString |

UPDATE_STR = "\x04\x02Æ\x9fÓ\x95\x0b\x00\x81«Íû´\x01\x0e\x01\x00\x0c\"\x91ÞûÕ\x08\x00¡«Íû´\x01\x17\x01\x00\x08¡\x91ÞûÕ\x08\x00\x01\x00\x02¡\x91ÞûÕ\x08\t\x01\x87Æ\x9fÓ\x95\x0b\x00\x06(\x00\x91ÞûÕ\x08\r\x06__type\x01w\x16recommendation-wrapper(\x00\x91ÞûÕ\x08\r\x08__format\x01}\x00(\x00\x91ÞûÕ\x08\r\x08__indent\x01}\x00!\x00\x91ÞûÕ\x08\r\x05__dir\x01(\x00\x91ÞûÕ\x08\r\x06__kind\x01w\x05write\x01\x00\x91ÞûÕ\x08\r\x01\x00\x02¨\x91ÞûÕ\x08\x0c\x01w\x03ltr¨\x91ÞûÕ\x08\x11\x01w\x03ltr\x81\x91ÞûÕ\x08\x13\x01\x00\x0e\x87\x91ÞûÕ\x08\x18\x06(\x00\x91ÞûÕ\x08'\x06__type\x01w\tparagraph(\x00\x91ÞûÕ\x08'\x08__format\x01}\x00(\x00\x91ÞûÕ\x08'\x08__indent\x01}\x00(\x00\x91ÞûÕ\x08'\x05__dir\x01w\x03ltr(\x00\x91ÞûÕ\x08'\x0c__textFormat\x01}\x00\x07\x00\x91ÞûÕ\x08'\x01(\x00\x91ÞûÕ\x08-\x06__type\x01w\x04text(\x00\x91ÞûÕ\x08-\x08__format\x01}\x00(\x00\x91ÞûÕ\x08-\x07__style\x01w\x00(\x00\x91ÞûÕ\x08-\x06__mode\x01}\x00(\x00\x91ÞûÕ\x08-\x08__detail\x01}\x00\x84\x91ÞûÕ\x08-\x01A\x87\x91ÞûÕ\x08'\x03\tUNDEFINED(\x00\x91ÞûÕ\x084\x06__type\x01w\x16recommendation-buttons(\x00\x91ÞûÕ\x084\x06__kind\x01w\x05write(\x00\x91ÞûÕ\x084\x14__recommendationDuid\x01w\x0casdf1234asdf\x07Á§\xadü\x04\x00!\x01\x04root\x05__dir\x01(\x01\x04root\x08__format\x01w\x00(\x01\x04root\x08__indent\x01}\x00(\x01\x04root\x06__type\x01w\x04root(\x01\x04root\t__version\x01}\x01\x01\x01\x04root\x01\x00\x06\x06«Íû´\x01\x00\x00\r¡Á§\xadü\x04\x00\x01\x81Á§\xadü\x04\x05\x01\x00\x08¡«Íû´\x01\r\x01\x00!\x04Á§\xadü\x04\x02\x00\x01\x05\x07\x91ÞûÕ\x08\x04\x00\r\x11\x01\x13\x03\x18\x0f«Íû´\x01\x01\x009Æ\x9fÓ\x95\x0b\x01\x00\r"
UPDATE_BYTES = [ord(e) for e in UPDATE_STR]


def test_lexical_parse_in_forward_direction():
    """This tests fully converting the Y format to JSON."""
    ydoc = Y.YDoc()
    Y.apply_update(ydoc, UPDATE_BYTES)
    yroot = ydoc.get_xml_fragment("root")

    result_json = {"root": yroot.to_dict()}

    print(f"{json.dumps(result_json, indent=4)}")

    assert EXPECTED_JSON == result_json


def test_lexical_parse_in_reverse_direction():
    """This tests fully converting the Y format to JSON."""
    ydoc = Y.YDoc()
    yroot = ydoc.get_xml_element("root")

    with ydoc.begin_transaction() as txn:
        nodes: list[tuple[Y.YXmlElement | Y.YXmlText | dict, dict]] = [
            (yroot, deepcopy(EXPECTED_JSON["root"]))
        ]
        while nodes:
            ynode, node_json = nodes.pop(0)
            for key, value in node_json.items():
                if key == "children" and isinstance(value, list):
                    for child in value:
                        if "text" in child:
                            text = child["text"]
                            del child["text"]
                            ynode.push_attributes(txn, child)
                            ynode.push(txn, text)
                        else:
                            ychild = ynode.push_xml_text(txn)
                            nodes.append((ychild, child))
                else:
                    ynode.set_attribute(txn, key, value)

    result_json = {"root": yroot.to_dict()}

    print(f"{json.dumps(result_json, indent=4)}")

    assert EXPECTED_JSON == result_json
